"""Creates the Windfarm class/model."""

import logging  # type: ignore
import os  # type: ignore

import networkx as nx  # type: ignore
import pandas as pd  # type: ignore
from geopy import distance  # type: ignore

from wombat.core import RepairManager, WombatEnvironment
from wombat.core.library import load_yaml
from wombat.windfarm.system import Cable, System


class Windfarm:
    def __init__(
        self,
        env: WombatEnvironment,
        windfarm_layout: str,
        repair_manager: RepairManager,
    ) -> None:

        self.env = env
        self.repair_manager = repair_manager

        # self._logging_setup()

        self._create_graph_layout(windfarm_layout)
        self._create_turbines_and_substations()
        self._create_cables()
        self.capacity = sum(self.node_system(turb).capacity for turb in self.turbine_id)

        self.env.process(self._log_operations())

    def _logging_setup(self) -> None:
        """Completes the setup for logging data.

        Parameters
        ----------
        which : str
            One of "events" or "operations". For creating event logs or operational
            status logs.
        """

        logging.basicConfig(
            format="%(asctime)s :: %(name)s :: %(levelname)s :: %(message)s",
            filename=self.env.operations_log_fname,
            filemode="w",
            level=logging.DEBUG,
        )
        self._operations_logger = logging.getLogger(__name__)

    def _create_graph_layout(self, windfarm_layout: str) -> None:
        """Creates a network layout of the windfarm start from the substation(s) to
        be able to capture downstream turbines that can be cut off in the event of a cable failure.

        Parameters
        ----------
        windfarm_layout : str
            Filename to use for reading in the windfarm layout; must be a csv file.
        """
        layout_path = os.path.join(self.env.data_dir, "windfarm", windfarm_layout)

        layout = (
            pd.read_csv(layout_path)
            .sort_values(by=["string", "order"])
            .reset_index(drop=True)
        )
        layout.subassembly = layout.subassembly.fillna("")

        windfarm = nx.DiGraph()
        windfarm.add_nodes_from(layout.id.values)

        # Assign the data attributes to the graph nodes
        for col in ("name", "latitude", "longitude", "subassembly"):
            d = {i: n for i, n in layout[["id", col]].values}
            nx.set_node_attributes(windfarm, d, name=col)

        # Determine which nodes are substations and which are turbines
        substation_filter = layout.id == layout.substation_id
        self.substation_id = layout[substation_filter].id.unique()
        self.turbine_id = layout[~substation_filter].id.unique()
        _type = {True: "substation", False: "turbine"}
        d = {i: _type[val] for i, val in zip(layout.id, substation_filter.values)}
        nx.set_node_attributes(windfarm, d, name="type")

        substations = layout[substation_filter].copy()
        turbines = layout[~substation_filter].copy()
        substation_sections = [
            turbines[turbines.substation_id == substation]
            for substation in substations.id
        ]
        for section in substation_sections:
            for _, row in section.iterrows():
                if row.order == 0:
                    start: str = row.substation_id
                else:
                    start = current  # noqa: F821
                current: str = row.id
                windfarm.add_edge(
                    start, current, length=row.distance, cable=row.upstream_cable
                )

        self.graph = windfarm

    def _create_turbines_and_substations(self) -> None:
        for system_id, data in self.graph.nodes(data=True):
            if data["subassembly"] == "":
                raise ValueError(
                    "A 'subassembly' file must be specified for all nodes in the windfarm layout!"
                )

            subassembly_dict = load_yaml(
                os.path.join(self.env.data_dir, "windfarm"), data["subassembly"]
            )
            self.graph.nodes[system_id]["system"] = System(
                self.env,
                self.repair_manager,
                system_id,
                data["name"],
                subassembly_dict,
                data["type"],
            )

    def _create_cables(self) -> None:
        for start_node, end_node, data in self.graph.edges(data=True):

            # If the real distance/cable length is not input, then the geodesic distance is calculated
            if data["length"] == 0:
                start_coordinates = (
                    self.graph.nodes[start_node]["latitude"],
                    self.graph.nodes[start_node]["longitude"],
                )
                end_coordinates = (
                    self.graph.nodes[end_node]["latitude"],
                    self.graph.nodes[end_node]["longitude"],
                )
                data["length"] = distance.geodesic(
                    start_coordinates, end_coordinates, ellipsoid="WGS-84"
                ).km

            upstream_turbines = nx.dfs_successors(self.graph, end_node)
            cable_dict = load_yaml(
                os.path.join(self.env.data_dir, "windfarm"), data["cable"]
            )
            data["cable"] = Cable(
                self,
                self.env,
                f"cable::{start_node}::{end_node}",
                start_node,
                upstream_turbines,
                cable_dict,
            )
        pass

    def _log_operations(self):
        """Logs the operational data for a simulation."""
        system_list = list(self.graph.nodes)
        columns = [
            "datetime",
            "name",
            "level",
            "env_datetime",
            "env_time",
        ] + system_list
        self.env._operations_logger.info(" :: ".join(columns))

        message = [self.env.simulation_time, self.env.now]
        message.extend(
            [
                self.graph.nodes[system]["system"].operating_level
                for system in system_list
            ]
        )
        message = " :: ".join((f"{m}" for m in message))
        self.env._operations_logger.info(message)

        HOURS = 1
        while True:
            yield self.env.timeout(HOURS)
            message = [self.env.simulation_time, self.env.now] + [
                self.node_system(system).operating_level for system in system_list
            ]
            message = " :: ".join(f"{m}" for m in message)
            self.env._operations_logger.info(message)

    def node_system(self, system_id: str) -> System:
        """Returns the desired node in the windfarm.

        Parameters
        ----------
        system_id : str
            The system's unique identifier, `wombat.windfarm.System.id`.

        Returns
        -------
        System
            The `System` object.
        """
        return self.graph.nodes[system_id]["system"]



"""Creates the RepairCrew class."""
# TODO: NEED A SPECIFIC STARTUP METHODs


import os  # type: ignore
from typing import Generator, List, Union  # type: ignore

import numpy as np  # type: ignore
from simpy.events import Process, Timeout  # type: ignore

from wombat.core import (
    Failure,
    Maintenance,
    RepairManager,
    RepairRequest,
    ServiceEquipmentData,
    WombatEnvironment,
)
from wombat.core.library import load_yaml
from wombat.utilities import hours_until_future_hour
from wombat.windfarm import Windfarm
from wombat.windfarm.system.cable import Cable
from wombat.windfarm.system.subassembly import Subassembly


HOURS_IN_DAY = 24


def consecutive_groups(data: np.ndarray, step_size: int = 1) -> List[np.ndarray]:
    """Generates the subgroups of an array where the difference between two sequential
    elements is equal to the `step_size`. The intent is to find the length of delays
    in a weather forecast

    Parameters
    ----------
    data : np.ndarray
        An array of integers.
    step_size : int, optional
        The step size to be considered a consecutive number, by default 1.

    Returns
    -------
    List[np.ndarray]
        A list of arrays of the consecutive elements of `data`.
    """
    # Consecutive groups in delay
    groups = np.split(data, np.where(np.diff(data) != step_size)[0] + 1)
    return groups


class ServiceEquipment:
    """Provides a servicing equipment object that can handle various maintenance and
    repair tasks.

    Parameters
    ----------
    env : WombatEnvironment
        The simulation environment.
    windfarm : Windfarm
        The `Windfarm` object.
    repair_manager : RepairManager
        The `RepairManager` object.
    equipment_data_file : str
        The equipment settings file name with extension.
    """

    def __init__(
        self,
        env: WombatEnvironment,
        windfarm: Windfarm,
        repair_manager: RepairManager,
        equipment_data_file: str,
    ):
        """Initializes the `ServiceEquipment` class.

        Parameters
        ----------
        env : WombatEnvironment
            The simulation environment.
        windfarm : Windfarm
            The `Windfarm` object.
        repair_manager : RepairManager
            The `RepairManager` object.
        equipment_data_file : str
            The equipment settings file name with extension.
        """
        self.env = env
        self.windfarm = windfarm
        self.manager = repair_manager

        data = load_yaml(
            os.path.join(env.data_dir, "repair", "transport"), equipment_data_file
        )
        if data["start_year"] < self.env.start_year:
            data["start_year"] = self.env.start_year
        if data["end_year"] > self.env.end_year:
            data["end_year"] = self.env.end_year
        self.settings = ServiceEquipmentData(**data)

        self.env.process(self.run())

    def _check_working_hours(self) -> None:
        """Checks the working hours of the equipment and overrides a default (-1) to
        the `env` settings, otherwise hours remain the same.
        """
        start_is_valid = self.settings.workday_start == -1
        end_is_valid = self.settings.workday_end == -1

        start = (
            self.env.workday_start if start_is_valid else self.settings.workday_start
        )
        end = self.env.workday_end if end_is_valid else self.settings.workday_end

        self.settings._set_environment_shift(start, end)

    def calculate_salary_cost(self, duration: Union[int, float]) -> float:
        """The total salaried labor cost implications for a given action.

        Parameters
        ----------
        duration : Union[int, float]
            Number of hours to charge for the salaried labor.

        Returns
        -------
        float
            The total cost of salaried workers for a given action.
        """
        return (
            (duration / HOURS_IN_DAY)
            * self.settings.day_rate
            * self.settings.n_day_rate
        )

    def calculate_hourly_cost(self, duration: Union[int, float]) -> float:
        """The total hourly labor cost implications for a given action.

        Parameters
        ----------
        duration : Union[int, float]
            Number of hours to charge for the hourly labor.

        Returns
        -------
        float
            The total cost of hourly workers for a given action.
        """
        return duration * self.settings.hourly_rate * self.settings.n_hourly_rate

    def calculate_equipment_cost(self, duration: Union[int, float]) -> float:
        """The total equipment cost implications for a given action.

        Parameters
        ----------
        duration : Union[int, float]
            Number of hours to charge for the equipment.

        Returns
        -------
        float
            The total cost of equipment for a given action.
        """
        return (duration / HOURS_IN_DAY) * self.settings.equipment_rate

    def register_repair_with_subassembly(
        self,
        subassembly: Union[Subassembly, Cable],
        repair: RepairRequest,
        starting_operating_level: float,
    ) -> None:
        """Goes into the repaired subassembly, component, or cable and returns its
        `operating_level` back to good as new for the specific repair. For fatal cable
        failures, all upstream turbines are turned back on unless there is another fatal
        cable failure preventing any more from operating.

        Parameters
        ----------
        subassembly : Union[Subassembly, Cable]
            The subassembly or cable that was repaired.
        repair : RepairRequest
            The request for repair that was submitted.
        starting_operating_level : float
            The operating level before a repair was started.
        """
        # If this is a cable repair and returning it back to operating, then reset all
        # upstream turbines to be operating as long as there isn't another fatal cable
        # failure preventing it from operating
        start: str = ""
        if repair.cable and repair.details.operation_reduction == 1:
            for i, turbine in enumerate(repair.upstream_turbines):
                if i > 0:
                    edge = self.windfarm.graph.edges[start, turbine]
                    if edge["cable"].operating_level == 0:
                        break
                self.windfarm.graph.nodes[turbine]["system"].cable_failure = False
                start = turbine

        # Put the subassembly/component back to good as new condition
        if repair.details.operation_reduction == 1:
            subassembly.operating_level = 1
            _ = self.manager.purge_subassembly_requests(
                repair.system_id, repair.subassembly_id
            )
        else:
            subassembly.operating_level += starting_operating_level
            subassembly.operating_level /= 1 - repair.details.operation_reduction

    def wait_until_next_shift(self, **kwargs) -> Generator[Timeout, None, None]:
        """Delays the crew until the start of the next shift.

        Yields
        -------
        Generator[Timeout, None, None]
            Delay until the start of the next shift.
        """
        delay = self.env.hours_to_next_shift(workday_start=self.settings.workday_start)
        salary_cost = self.calculate_salary_cost(delay)
        hourly_cost = self.calculate_hourly_cost(0)
        equpipment_cost = self.calculate_equipment_cost(delay)
        self.env.log_action(
            system_id=kwargs.get("system_id", ""),
            system_name=kwargs.get("system_name", ""),
            part_id=kwargs.get("part_id", ""),
            part_name=kwargs.get("part_name", ""),
            system_ol=kwargs.get("system_ol", ""),
            part_ol=kwargs.get("part_ol", ""),
            agent=kwargs.get("agent", ""),
            action=kwargs.get("action", "delay"),
            reason=kwargs.get("reason", ""),
            additional=kwargs.get(
                "additional", "work shift has ended; waiting for next shift to start"
            ),
            duration=delay,
            request_id=kwargs.get("request_id", "na"),
            salary_labor_cost=salary_cost,
            hourly_labor_cost=hourly_cost,
            equipment_cost=equpipment_cost,
        )
        yield self.env.timeout(delay)

    def wait_until_next_operational_period(self) -> Generator[Timeout, None, None]:
        """Delays the crew and equipment until the start of the next operational
        period.

        TODO: Need a custom error if weather doesn't align with the equipment dates.

        Yields
        -------
        Generator[Timeout, None, None]
            A Timeout event for the number of hours between when the function is called
            and when the next operational period starts.
        """
        current = self.env.simulation_time.date()
        ix_match = np.where(current < self.settings.operating_dates)[0]
        if ix_match.size > 0:
            still_in_operation = True
            next_operating_date = self.settings.operating_dates[ix_match[0]]
            hours_to_next_shift = (
                self.env.date_ix(next_operating_date)
                - self.env.now
                + self.env.workday_start
            )
        else:
            still_in_operation = False
            hours_to_next_shift = self.env.max_run_time - self.env.now

        if still_in_operation:
            additional = "will return next year"
        else:
            additional = "no more return visits will be made"

        self.env.log_action(
            system_id="",
            system_name="",
            part_id="",
            part_name="",
            system_ol="",
            part_ol="",
            agent=self.settings.name,
            action="delay",
            reason="work is complete",
            additional=additional,
            duration=0,
            salary_labor_cost=0,
            hourly_labor_cost=0,
            equipment_cost=0,
        )

        yield self.env.timeout(hours_to_next_shift)

    def mobilize(self) -> Generator[Timeout, None, None]:
        """Mobilizes the Equipment object by waiting for the next operational period,
        then logging the mobiliztion cost.

        Yields
        -------
        Generator[Timeout, None, None]
            A Timeout event for the number of hours between when the function is called
            and when the next operational period starts.
        """
        yield self.env.process(self.wait_until_next_operational_period())

        self.env.log_action(
            system_id="",
            system_name="",
            part_id="",
            part_name="",
            system_ol="",
            part_ol="",
            agent=self.settings.name,
            action="mobilization",
            reason=f"{self.settings.name} is scheduled",
            additional="mobilization",
            duration=0,
            salary_labor_cost=0,
            hourly_labor_cost=0,
            equipment_cost=self.settings.mobilization_cost,
        )

    def assess_weather(
        self, hours_required: Union[int, float], n_tries: int = 0
    ) -> np.ndarray:
        """Assesses at which points in the repair window, the wind (and wave)
        constraints for safe operation are met. If there is a delay in the time window,
        then the a new window with the added delay is processed to ensure that an array
        with enough safe operating times is available within the returned array.

        Parameters
        ----------
        hours_required : Union[int, float]
            The number of hours required to complete the repair.
        n_tries : int
            The number of previous attempts to find a suitable weather window.

        Returns
        -------
        np.ndarray
            A boolean array for what points in the time window are safe to complete a
            maintenance task or repair.
        """
        hours_required = np.ceil(hours_required).astype(int)
        wind, wave = self.env.weather_forecast(hours_required).T  # type: ignore
        satisfactory_wind = wind <= self.settings.max_windspeed_repair
        satisfactory_wave = wave <= self.settings.max_waveheight_repair
        all_clear = satisfactory_wind & satisfactory_wave

        delay = all_clear.size - np.count_nonzero(all_clear)
        enough_windows = hours_required <= np.count_nonzero(all_clear)
        shift_exceeded = all_clear.size > self.env.shift_length
        if enough_windows or shift_exceeded:
            return all_clear

        # TODO: Need a better solution to for the recursion error
        if n_tries < 10:
            return self.assess_weather(hours_required + delay, n_tries=n_tries + 1)
        else:
            return all_clear

    def get_next_request(self):
        """Gets the next request by the rig's method for processing repairs.

        Returns
        -------
        simpy.resources.store.FilterStoreGet
            The next `RepairRequest` to be processed.
        """
        if self.settings.method == "turbine":
            return self.manager.get_request_by_turbine(self.settings.capability)
        elif self.settings.method == "severity":
            return self.manager.get_next_highest_severity_request(
                self.settings.capability
            )

    def process_repair(
        self,
        request: RepairRequest,
        time_processed: Union[int, float] = 0,
        prior_operation_level: float = -1.0,
    ) -> Generator[Union[Timeout, Process], None, None]:
        """Processes the repair including any weather and shift delays.

        Parameters
        ----------
        request : RepairRequest
            The `Maintenance` or `Failure` receiving attention.
        time_processed : Union[int, float], optional
            Time that has already been processed, by default 0.
        prior_operation_level : float, optional
            The operating level of the `System` just before the repair has begun, by
            default -1.0.

        Yields
        -------
        Generator[Union[Timeout, Process], None, None]
            Timeouts for the repair process.
        """
        shift_delay = False
        hours_required = request.details.time - time_processed
        system = self.windfarm.graph.nodes[request.system_id]["system"]
        if request.cable:
            cable = request.subassembly_id.split("::")[1:]
            subassembly = self.windfarm.graph.edges[cable]["cable"]
        else:
            subassembly = getattr(system, request.subassembly_id)

        staring_operational_level = max(
            prior_operation_level, subassembly.operating_level
        )

        # Obtain the weather forecast for the duration of the required time
        weather_forecast = self.assess_weather(hours_required + 1)

        # Check that all times are within the windfarm's working hours and cut off any
        # time points that fall outside of the work shifts
        current = self.env.simulation_time
        time_in_work_shift = np.arange(
            current.hour, current.hour + weather_forecast.size
        )
        working_hours = np.array(
            [
                self.env.hour_in_shift(
                    hour,
                    workday_start=self.settings.workday_start,
                    workday_end=self.settings.workday_end,
                )
                for hour in time_in_work_shift
            ]
        )
        if np.any(~working_hours):
            shift_delay = True
            hours_until_end = np.where(~working_hours)[0][0]
            weather_forecast = weather_forecast[:hours_until_end]

        # First turn off the turbine, then proceed with the repair or maintenance so the
        # turbine is not registered as operating when the turbine is being worked on
        subassembly.operating_level = 0
        subassembly.interrupt_all_subassembly_processes()

        shared_logging = dict(
            system_id=system.id,
            system_name=system.name,
            part_id=subassembly.id,
            part_name=subassembly.name,
            system_ol=system.operating_level,
            part_ol=subassembly.operating_level,
            agent=self.settings.name,
            reason=request.details.description,
            request_id=request.request_id,
        )
        hours_processed = 0
        delay_groups = consecutive_groups(np.where(~weather_forecast)[0])
        while weather_forecast.size > 0 and hours_required > 0:
            delays = np.where(~weather_forecast)[0]
            if delays.size > 0:
                hours_to_process = delays[0]
                delay = delay_groups.pop(0).size
            else:
                hours_to_process = weather_forecast.size
                delay = 0
            weather_forecast = weather_forecast[hours_to_process + delay :]

            if hours_to_process > 0:
                hours_to_process = int(min(hours_required, hours_to_process))
                current = self.env.simulation_time
                hours_to_process = hours_until_future_hour(
                    current, current.hour + hours_to_process
                )

                action = (
                    "repair" if isinstance(request.details, Failure) else "maintenance"
                )
                salary_cost = self.calculate_salary_cost(hours_to_process)
                hourly_cost = self.calculate_hourly_cost(hours_to_process)
                equipment_cost = self.calculate_equipment_cost(hours_to_process)
                self.env.log_action(
                    action=action,
                    duration=hours_to_process,
                    salary_labor_cost=salary_cost,
                    hourly_labor_cost=hourly_cost,
                    equipment_cost=equipment_cost,
                    additional=action,
                    **shared_logging,
                )
                yield self.env.timeout(hours_to_process)

                hours_processed += hours_to_process
                hours_required -= hours_to_process

            if delay > 0:
                current = self.env.simulation_time
                hours_to_process = hours_until_future_hour(
                    current, current.hour + delay
                )

                salary_cost = self.calculate_salary_cost(hours_to_process)
                hourly_cost = self.calculate_hourly_cost(0)
                equipment_cost = self.calculate_equipment_cost(hours_to_process)
                self.env.log_action(
                    action="delay",
                    additional="weather delay",
                    duration=hours_to_process,
                    salary_labor_cost=salary_cost,
                    hourly_labor_cost=hourly_cost,
                    equipment_cost=equipment_cost,
                    **shared_logging,
                )
                yield self.env.timeout(hours_to_process)

        if shift_delay:
            yield self.env.process(self.wait_until_next_shift(**shared_logging))

            yield self.env.process(
                self.process_repair(
                    request,
                    time_processed=hours_processed + time_processed,
                    prior_operation_level=staring_operational_level,
                )
            )
            return

        # Register the repair
        self.register_repair_with_subassembly(
            subassembly, request, staring_operational_level
        )

        action = "maintenance" if isinstance(request.details, Maintenance) else "repair"
        self.env.log_action(
            system_id=system.id,
            system_name=system.name,
            part_id=subassembly.id,
            part_name=subassembly.name,
            system_ol=system.operating_level,
            part_ol=subassembly.operating_level,
            agent=self.settings.name,
            action=f"{action} complete",
            reason=request.details.description,
            materials_cost=request.details.materials,
            additional="complete",
            duration=0,
            request_id=request.request_id,
        )

    def run(self) -> Generator[Process, None, None]:
        """Runs the simulation.

        Yields
        -------
        Generator[Process, None, None]
            The simulation.
        """
        while True:
            # Wait for a valid operational period to start
            if self.env.simulation_time.date() not in self.settings.operating_dates:

                yield self.env.process(self.mobilize())

            # Wait for next shift to start
            is_workshift = self.env.is_workshift(
                workday_start=self.settings.workday_start,
                workday_end=self.settings.workday_end,
            )
            if not is_workshift:
                yield self.env.process(
                    self.wait_until_next_shift(
                        **dict(
                            agent=self.settings.name,
                            reason="not in working hours",
                            additional="not in working hours",
                        )
                    )
                )

            request = self.get_next_request()
            if request is None:
                yield self.env.process(
                    self.wait_until_next_shift(
                        **dict(
                            agent=self.settings.name,
                            reason="no requests",
                            additional="no work requests submitted by start of shift",
                        )
                    )
                )
            else:
                yield self.env.process(self.process_repair(request.value))



"""Creates the necessary repair classes."""


from typing import Iterator, Optional, Sequence  # type: ignore

import numpy as np  # type: ignore
from simpy.resources.store import FilterStore, FilterStoreGet  # type: ignore

from wombat.core import Failure, Maintenance, RepairRequest, WombatEnvironment


class RepairManager(FilterStore):
    """Provides a class to manage repair and maintenance tasks.

    Parameters
    ----------
    FilterStore : simpy.resources.store.FilterStore
        The `simpy` class on which RepairManager is based to manage the repair and
        maintenance tasks.
    env : wombat.core.WombatEnvironment
        The simulation environment.
    capacity : float
        The maximum number of tasks that can be submitted to the manager, by default `np.inf`.
    """

    def __init__(self, env: WombatEnvironment, capacity: float = np.inf) -> None:
        """Initializes the class.

        Parameters
        ----------
        env : WombatEnvironment
            The simualation environment.
        capacity : float, optional
            The maximum number of tasks that the manager should be storing, by default
            `np.inf`.
        """
        super().__init__(env, capacity)
        self.env = env
        self._current_id = 0

    def _create_request_id(self, request: RepairRequest) -> str:
        """Creates a unique `request_id` to be logged in the `request`.

        Parameters
        ----------
        request : RepairRequest
            The request object.

        Returns
        -------
        str
            An 11-digit identifier starting with "MNT" for maintenance tasks or "RPR"
            for repairs.

        Raises
        ------
        ValueError
            If the `request.details` property is not a `Failure` or `Maintenance` object,
            then a ValueError will be raised.
        """
        if isinstance(request.details, Failure):
            prefix = "RPR"
        elif isinstance(request.details, Maintenance):
            prefix = "MNT"
        else:
            raise ValueError("A valid `RepairRequest` must be submitted")

        request_id = f"{prefix}{str(self._current_id).zfill(8)}"
        self._current_id += 1
        return request_id

    def submit_request(self, request: RepairRequest) -> RepairRequest:
        """The method to submit requests to the repair mananger and adds a unique
        identifier for logging.

        Parameters
        ----------
        request : RepairRequest
            The request that needs to be tracked and logged.

        Returns
        -------
        RepairRequest
            The same request as passed into the method, but with a unique identifier
            used for logging.
        """
        request_id = self._create_request_id(request)
        request.assign_id(request_id)
        self.put(request)
        return request

    def get_request_by_turbine(
        self, equipment_capability: Sequence[str], turbine_id: Optional[str] = None
    ) -> Optional[FilterStoreGet]:
        """Gets all repair requests for a certain turbine with maximum severity of `severity_max`.

        Note: For now, this will simply order the items by `Turbine.id` and
        return the first turbine.

        Parameters
        ----------
        equipment_capability : List[str]
            The capability of the equipment requesting possible repairs to make.
        turbine_id : Optional[str], optional
            ID of the turbine; should correspond to `Turbine.id`, by default None.
            If None, then it will simply sort the list by `Turbine.id` and
            return the first repair requested.

        Returns
        -------
        Optional[FilterStoreGet]
            All of the repair requests for the focal turbine or None, if self.items
            is empty.
        """
        if not self.items:
            return None

        equipment_capability = set(equipment_capability)  # type: ignore
        for request in sorted(self.items, key=lambda x: x.turbine_id):
            capability_overlap = equipment_capability.intersection(request.equipment)  # type: ignore
            valid_turbine = turbine_id is None or request.turbine_id == turbine_id
            if valid_turbine and capability_overlap:
                return self.get(lambda x: x == request)

        # In case the loop above iterates all the way to the end, nothing was
        # found so return None. This is probably an error for which an error
        # should be raised, but this keeps the type hints correct for now.

        return None

    def get_next_highest_severity_request(
        self, equipment_capability: Sequence[str], severity_level: Optional[int] = None
    ) -> Optional[FilterStoreGet]:
        """Gets the next repair request by `severity_level`.

        Parameters
        ----------
        equipment_capability : List[str]
            The capability of the equipment requesting possible repairs to make.
        severity_level : int
            Severity level of focus, default None.

        Returns
        -------
        Optional[FilterStoreGet]
            Repair request meeting the required criteria.
        """
        if not self.items:
            return None

        equipment_capability = set(equipment_capability)  # type: ignore
        requests = sorted(self.items, key=lambda x: x.severity_level, reverse=True)
        for request in requests:
            capability_overlap = equipment_capability & set(  # type: ignore
                request.details.service_equipment
            )
            requested_severity: bool = severity_level in (request.details.level, None)
            if requested_severity and capability_overlap:
                return self.get(lambda x: x == request)

        # This return statement ensures there is always a known return type,
        # but consider raising an error here if the loop is always assumed to
        # break when something is found.

        return None

    def purge_subassembly_requests(
        self, system_id: str, subassembly_id: str
    ) -> Optional[Iterator]:
        """Yields all the requests for a system/subassembly combination. This is intended
        to be used to remove erroneous requests after a subassembly has been replaced.

        Parameters
        ----------
        system_id : str
            `System.id`.
        subassembly_id : str
            `Subassembly.id`.

        Yields
        -------
        Optional[Iterator]
            All requests made to the repair manager for the provided system/subassembly
            combination. Returns None if self.items is empty or the loop terminates
            without finding what it is looking for.
        """
        if not self.items:
            return None

        for request in self.items:
            system_match = request.system_id == system_id
            subassembly_match = request.subassembly_id == subassembly_id
            if system_match and subassembly_match:
                yield self.get(lambda x: x == request)

                self.env.log_action(
                    system_id=request.system_id,
                    system_name=request.system_name,
                    part_id=request.system_name,
                    part_name=request.part_name,
                    system_ol="n/a",
                    part_ol="n/a",
                    agent="RepairManager",
                    action="request canceled",
                    reason="replacement required",
                    request_id=request.request_id,
                )

        # This is here to ensure there is always an explicit return value,
        # but you may want to consider raising an error here.

        return None



"""The main API for the `wombat`."""

from pathlib import Path
from typing import List, Optional, Union

import attr
import pandas as pd
from simpy.events import Event  # type: ignore

from wombat.core import Metrics, RepairManager, ServiceEquipment, WombatEnvironment
from wombat.core.library import library_map, load_yaml
from wombat.windfarm import Windfarm


def _library_mapper(file_path: Union[str, Path]) -> Union[str, Path]:
    """Attempts to extract a default library path if one of "DINWOODIE" or "IEA_26"
    are passed, other returns `file_path`.

    Parameters
    ----------
    file_path : Union[str, Path]
        Should be a valid file path, or one of "DINWOODIE" or "IEA_26" to indicate a
        provided library is being used.

    Returns
    -------
    Union[str, Path]
        The library path.
    """
    return library_map.get(file_path, file_path)


@attr.s(frozen=True, auto_attribs=True)
class Configuration:
    """The `Simulation` configuration data class that provides all the necessary definitions.

    Parameters
    ----------
    name: str
        Name of the simulation. Used for logging files.
    library : str
        The data directory. See `wombat.simulation.WombatEnvironment` for more details.
    layout : str
        The windfarm layout file. See `wombat.Windfarm` for more details.
    service_equipment : Union[str, List[str]]
        The equpiment that will be used in the simulation. See
        `wombat.core.ServiceEquipment` for more details.
    weather : str
        The weather profile to be used. See `wombat.simulation.WombatEnvironment`
        for more details.
    workday_start : int
        Starting hour for a typical work shift. Can be overridden by
        equipment-specific settings.
    workday_end : int
        Ending hour for a typical work shift. Can be overridden by
        equipment-specific settings.
    inflation_rate : float
        The annual inflation rate to be used for post-processing.
    fixed_costs : str
        The file name for the fixed costs assumptions.
    project_capacity : Union[int, float]
        The total capacity of the wind plant, in MW.
    start_year : int
        Start year of the simulation. The exact date will be determined by
        the first valid date of this year in `weather`.
    end_year : int
        Final year of the simulation. The exact date will be determined by
        the last valid date of this year in `weather`.
    SAM_settings : str
        The SAM settings file to be used for financial modeling, optional, by
        default None.
    """

    name: str
    library: Path
    layout: str = attr.ib(converter=_library_mapper)
    service_equipment: Union[str, List[str]]
    weather: Union[str, pd.DataFrame]
    workday_start: int
    workday_end: int
    inflation_rate: float
    fixed_costs: str
    project_capacity: Union[int, float]
    start_year: int = attr.ib(default=None)
    end_year: int = attr.ib(default=None)
    SAM_settings: str = attr.ib(default=None)

    def __attrs_post_init__(self):
        if isinstance(self.service_equipment, str):
            object.__setattr__(self, "service_equipment", [self.service_equipment])


class Simulation:
    """The primary API to interact with the simulation methodologies.

    Parameters
    ----------
    name: str
        Name of the simulation. Used for logging files.
    library_path : str
        The path to the main data library.
    config : Union[str, dict]
        The path to a configuration dictionary or the dictionary itself.
    """

    def __init__(self, name: str, library_path: str, config: Union[str, dict]):
        """Creates a `Simulation` object with an analysis ready to run.

        Parameters
        ----------
        name: str
            Name of the simulation. Used for logging files.
        library_path : str
            The path to the main data library. If one of DINWOODIE or IEA_26
        config : Union[str, dict]
            The path to a configuration dictionary or the dictionary itself.

        Returns
        -------
        Simulation
            A `Simulation` object
        """
        library_path = _library_mapper(library_path)
        self.data_dir = Path(library_path).resolve()
        if isinstance(config, str):
            config = load_yaml(self.data_dir / "config", config)  # type: ignore
        config["name"] = name  # type: ignore
        config["library"] = self.data_dir  # type: ignore
        self.config = Configuration(**config)  # type: ignore

        self.setup_simulation()

    @classmethod
    def from_inputs(
        cls,
        name: str,
        library: str,
        layout: str,
        service_equipment: Union[str, List[str]],
        weather: str,
        start_year: int,
        end_year: int,
        workday_start: int,
        workday_end: int,
        inflation_rate: float,
        fixed_costs: str,
        project_capacity: Union[int, float],
        SAM_settings: str = None,
    ):
        """Creates a `Simulation` object from a series of inputs instead of a
        predefined dictionary or YAML file.

        Parameters
        ----------
        name: str
            Name of the simulation. Used for logging files.
        library : str
            The data directory. See `wombat.simulation.WombatEnvironment` for more details.
        layout : str
            The windfarm layout file. See `wombat.Windfarm` for more details.
        service_equipment : Union[str, List[str]]
            The equpiment that will be used in the simulation. See
            `wombat.core.ServiceEquipment` for more details.
        weather : str
            The weather profile to be used. See `wombat.simulation.WombatEnvironment`
            for more details.
        start_year : int
            Start year of the simulation. The exact date will be determined by
            the first valid date of this year in `weather`.
        end_year : int
            Final year of the simulation. The exact date will be determined by
            the last valid date of this year in `weather`.
        workday_start : int
            Starting hour for a typical work shift. Can be overridden by
            equipment-specific settings.
        workday_end : int
            Ending hour for a typical work shift. Can be overridden by
            equipment-specific settings.
        inflation_rate : float
            The annual inflation rate to be used for post-processing.
        fixed_costs : str
            The file name for the fixed costs assumptions.
        project_capacity : Union[int, float]
            The total capacity of the wind plant, in MW.
        SAM_settings : str
            The SAM settings file to be used for financial modeling, optional, by
            default None.

        Returns
        -------
        Simulation
            Returns a `Simulation` object.
        """
        config = dict(
            name=name,
            library=_library_mapper(library),
            layout=layout,
            service_equipment=service_equipment,
            weather=weather,
            start_year=start_year,
            end_year=end_year,
            workday_start=workday_start,
            workday_end=workday_end,
            inflation_rate=inflation_rate,
            fixed_costs=fixed_costs,
            project_capacity=project_capacity,
            SAM_settings=SAM_settings,
        )
        return cls(name, library, config)

    def setup_simulation(self):
        """Initializes the simulation objects."""
        self.env = WombatEnvironment(
            self.config.library,
            self.config.weather,
            simulation_name=self.config.name,
            workday_start=self.config.workday_start,
            workday_end=self.config.workday_end,
            start_year=self.config.start_year,
            end_year=self.config.end_year,
        )
        self.repair_manager = RepairManager(self.env)
        self.windfarm = Windfarm(self.env, self.config.layout, self.repair_manager)
        self.service_equipment = []
        for service_equipment in self.config.service_equipment:
            self.service_equipment.append(
                ServiceEquipment(
                    self.env, self.windfarm, self.repair_manager, service_equipment
                )
            )

    def run(self, until: Optional[Union[int, float, Event]] = None):
        """Calls `WombatEnvironment.run()` and gathers the results for post-processing.
        See `wombat.simulation.WombatEnvironment.run` or `simpy.Environment.run` for more
        details.

        Parameters
        ----------
        until : Optional[Union[int, float, Event]], optional
            When to stop the simulation, by default None. See documentation on
            `simpy.Environment.run` for more details.
        """
        self.env.run(until=until)

        operations, events = self.env.convert_logs_to_csv(return_df=True)
        power_potential, power_production = self.env.power_production_potential_to_csv(
            windfarm=self.windfarm, operations=operations, return_df=True
        )
        self.metrics = Metrics(
            self.config.library,
            events,
            operations,
            power_potential,
            power_production,
            self.config.inflation_rate,
            self.config.project_capacity,
            [
                self.windfarm.node_system(t_id).capacity
                for t_id in self.windfarm.turbine_id
            ],
            self.config.fixed_costs,
            self.windfarm.substation_id.tolist(),
            self.windfarm.turbine_id.tolist(),
            [el.settings.name for el in self.service_equipment],
            SAM_settings=self.config.SAM_settings,
        )



"""Turbine and turbine component shared utilities."""


import datetime  # type: ignore
from math import fsum  # type: ignore
from typing import Callable, Dict, List, Sequence, Union  # type: ignore

import attr  # type: ignore
import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from scipy.stats import weibull_min  # type: ignore


HOURS_IN_YEAR = 8760
HOURS_IN_DAY = 24


def convert_to_list(
    value: Union[Sequence, Union[str, int, float]], manipulation: Callable = None
):
    """Converts an unknown element that could be a list or single, non-sequence element
    to a list of elements.

    Parameters
    ----------
    value : Union[Sequence, Union[str, int, float]]
        The unknown element to be converted to a list of element(s).
    manipulation: Callable
        A function to be performed upon the individual elements

    Returns
    -------
    List[Any]
        The new list of elements.
    """

    if isinstance(value, (str, int, float)):
        value = [value]
    if manipulation is not None:
        return [manipulation(el) for el in value]
    return list(value)


def annual_date_range(
    start_day: int,
    end_day: int,
    start_month: int,
    end_month: int,
    start_year: int,
    end_year: int,
) -> np.ndarray:
    """Creates a series of date ranges across multiple years.

    Parameters
    ----------
    start_day : int
        Starting day of the month.
    end_day : int
        Ending day of the month.
    start_month : int
        Starting month.
    end_month : int
        Ending month.
    start_year : int
        First year of the date range.
    end_year : int
        Last year of the date range.

    Yields
    -------
    np.ndarray
        A `numpy.ndarray` of `datetime.date` objects.
    """
    date_ranges = []
    if end_month >= start_month:
        for year in range(start_year, end_year + 1):
            start = datetime.datetime(year, start_month, start_day)
            end = datetime.datetime(year, end_month, end_day)
            date_ranges.extend(pd.date_range(start, end).date)
    else:
        for year in range(start_year, end_year + 1):
            start1 = datetime.datetime(year, start_month, start_day)
            end1 = datetime.datetime(year, 12, 31)
            date_ranges.extend(pd.date_range(start1, end1).date)
            if year != end_year:
                start2 = datetime.datetime(year + 1, 1, 1)
                end2 = datetime.datetime(year + 1, end_month, end_day)
                date_ranges.extend(pd.date_range(start2, end2).date)

    date_range = np.array(date_ranges)
    return date_range


def convert_ratio_to_absolute(
    ratio: Union[int, float], total: Union[int, float]
) -> Union[int, float]:
    """Converts a proportional cost to an absolute cost.

    Parameters
    ----------
    ratio : Union[int, float]
        The proportional materials cost for a `Maintenance` or `Failure`.
    total: Union[int, float]
        The turbine's replacement cost.

    Returns
    -------
    Union[int, float]
        The absolute cost of the materials.
    """

    if ratio <= 1:
        return ratio * total
    return ratio


@attr.s(frozen=True, auto_attribs=True)
class Maintenance:
    """Data class to store maintenance data used in subassembly and cable modeling.

    Parameters
    ----------
    time : float
        Amount of time required to perform maintenance, in hours.
    materials : float
        Cost of materials required to perform maintenance, in USD.
    frequency : float
        Optimal number of days between performing maintenance, in days.
    service_equipment: Union[List[str], str]
        Any combination of the `Equipment.capability` options.
         - RMT: remote (no actual equipment BUT no special implementation)
         - DRN: drone
         - CTV: crew transfer vessel/vehicle
         - SCN: small crane (i.e., field support vessel)
         - LCN: large crane (i.e., heavy lift vessel)
         - CAB: cabling vessel/vehicle
         - DSV: diving support vessel
    system_value : Union[int, float]
        Turbine replacement value. Used if the materials cost is a proportional cost.
    description : str
        A short text description to be used for logging.
    """

    time: float
    materials: float
    frequency: float
    service_equipment: Union[List[str], str]
    system_value: Union[int, float]
    description: str = "routine maintenance"
    level: int = 0
    operation_reduction: float = 0.0
    request_id: str = attr.ib(init=False)

    def __attrs_post_init__(self):
        """Convert frequency to hours (simulation time scale) and the equipment
        requirement to a list.
        """
        object.__setattr__(self, "frequency", self.frequency * HOURS_IN_DAY)
        object.__setattr__(
            self,
            "service_equipment",
            convert_to_list(self.service_equipment, str.upper),
        )
        object.__setattr__(
            self,
            "materials",
            convert_ratio_to_absolute(self.materials, self.system_value),
        )

    def assign_id(self, request_id: str) -> None:
        """Assigns a unique identifier to the request.

        Parameters
        ----------
        request_id : str
            The `wombat.core.RepairManager` generated identifier.
        """
        object.__setattr__(self, "request_id", request_id)


@attr.s(frozen=True, auto_attribs=True)
class Failure:
    """Data class to store failure data used in subassembly and cable modeling.

    Parameters
    ----------
    scale : float
        Weibull scale parameter for a failure classification.
    shape : float
        Weibull shape parameter for a failure classification.
    time : float
        Amount of time required to complete the repair, in hours.
    materials : float
        Cost of the materials required to complete the repair, in $USD.
    operation_reduction : float
        Performance reduction caused by the failure, between (0, 1].
    level : int, optional
        Level of severity, will be generated in the `ComponentData.create_severities`
        method.
    service_equipment: Union[List[str], str]
        Any combination of the `Equipment.capability` options.
         - RMT: remote (no actual equipment BUT no special implementation)
         - DRN: drone
         - CTV: crew transfer vessel/vehicle
         - SCN: small crane (i.e., field support vessel)
         - LCN: large crane (i.e., heavy lift vessel)
         - CAB: cabling vessel/vehicle
         - DSV: diving support vessel
    system_value : Union[int, float]
        Turbine replacement value. Used if the materials cost is a proportional cost.
    description : str
        A short text description to be used for logging.
    """

    scale: float
    shape: float
    time: float
    materials: float
    operation_reduction: float
    level: int
    service_equipment: Union[List[str], str]
    system_value: Union[int, float]
    description: str = "failure"
    weibull: weibull_min = attr.ib(init=False)
    request_id: str = attr.ib(init=False)

    def __attrs_post_init__(self):
        """Creates the actual Weibull distribution and converts equipment requirements
        to a list.
        """
        object.__setattr__(self, "weibull", weibull_min(self.shape, scale=self.scale))
        object.__setattr__(
            self,
            "service_equipment",
            convert_to_list(self.service_equipment, str.upper),
        )
        object.__setattr__(
            self,
            "materials",
            convert_ratio_to_absolute(self.materials, self.system_value),
        )

    def hours_to_next_failure(self) -> Union[None, float]:
        """Samples the next time to failure in a Weibull distribution. If the `scale`
        and `shape` parameters are set to 0, then the model will return `None` to cause
        the subassembly to timeout to the end of the simulation.

        Returns
        -------
        Union[None, float]
            Returns `None` for a non-modelled failure, or the time until the next
            simulated failure.
        """
        if self.scale == self.shape == 0:
            return None

        return self.weibull.rvs(size=1)[0] * HOURS_IN_YEAR

    def assign_id(self, request_id: str) -> None:
        """Assigns a unique identifier to the request.

        Parameters
        ----------
        request_id : str
            The `wombat.core.RepairManager` generated identifier.
        """
        object.__setattr__(self, "request_id", request_id)


@attr.s(frozen=True, auto_attribs=True)
class SubassemblyData:
    """Data storage and validation class for the subassemblies.

    Parameters
    ----------
    name : str
        Name of the component/subassembly.
    maintenance : List[Dict[str, Union[float, str]]]
        List of the maintenance classification dictionaries. This will be converted
        to a list of `Maintenance` objects in the post initialization hook.
    failures : Dict[int, Dict[str, Union[float, str]]]
        Dictionary of failure classifications in a numerical (ordinal) categorization
        order. This will be converted to a dictionary of `Failure` objects in the
        post initialization hook.
    system_value : Union[int, float]
        Turbine's cost of replacement. Used in case percentages of turbine cost are used
        in place of an absolute cost.
    """

    name: str
    maintenance: List[Dict[str, Union[float, str]]]
    failures: Dict[int, Dict[str, Union[float, str]]]
    system_value: Union[int, float]

    def __attrs_post_init__(self):
        """Converts the maintenance and failure data to `Maintenance` and `Failure` objects, respectively."""
        object.__setattr__(
            self,
            "maintenance",
            [
                Maintenance(**el, **{"system_value": self.system_value})
                for el in self.maintenance
            ],
        )
        object.__setattr__(
            self,
            "failures",
            {
                level: Failure(**values, **{"system_value": self.system_value})
                for level, values in self.failures.items()
            },
        )


@attr.s(frozen=True, auto_attribs=True)
class RepairRequest:
    """Repair/Maintenance request data class.

    Parameters
    ----------
    system_id : str
        `System.id`.
    system_name : str
        `System.name`.
    subassembly_id : str
        `Subassembly.id`.
    subassembly_name : str
        `Subassembly.name`.
    severity_level : int
        `Maintenance.level` or `Failure.level`.
    details : Union[Failure, Maintenance]
        The actual data class.
    cable : bool
        Indicator that the request is for a cable, by default False.
    upstream_turbines : List[str]
        The cable's upstream turbines, by default []. No need to use this if `cable` == False.
    """

    system_id: str
    system_name: str
    subassembly_id: str
    subassembly_name: str
    severity_level: int
    details: Union[Failure, Maintenance]
    cable: bool = attr.ib(default=False)
    upstream_turbines: List[str] = attr.Factory(list)
    request_id: str = attr.ib(default="")

    def assign_id(self, request_id: str) -> None:
        """Assigns a unique identifier to the request.

        Parameters
        ----------
        request_id : str
            The `wombat.core.RepairManager` generated identifier.
        """
        object.__setattr__(self, "request_id", request_id)
        self.details.assign_id(request_id)


@attr.s(frozen=True, auto_attribs=True)
class ServiceEquipmentData:
    """Crew data class.

    Parameters
    ----------
    name: str
        Name of the piece of servicing equipment.
    equipment_rate: float
        Day rate for the equipment/vessel, in USD.
    day_rate: float
        Day rate for salaried workers, in USD.
    n_day_rate: int
        Number of salaried workers.
    hourly_rate: float
        Hourly labor rate for subcontractors, in USD.
    n_hourly_rate: int
        Number of hourly/subcontractor workers.
    start_month : int
        The day to start operations for the rig and crew.
    start_day : int
        The month to start operations for the rig and crew.
    start_year : int
        The year to start operations for the rig and crew.
    end_month : int
        The month to end operations for the rig and crew.
    end_day : int
        The day to end operations for the rig and crew.
    end_year : int
        The year to end operations for the rig and crew.
        ... note:: if the rig comes annually, then the enter the year for the last year
        that the rig and crew will be available.
    capability : str
        The type of capabilities the equipment contains. Must be one of:
         - RMT: remote (no actual equipment BUT no special implementation)
         - DRN: drone
         - CTV: crew transfer vessel/vehicle
         - SCN: small crane (i.e., field support vessel)
         - LCN: large crane (i.e., heavy lift vessel)
         - CAB: cabling vessel/vehicle
         - DSV: diving support vessel
    mobilization_cost : float
        Cost to mobilize the rig and crew.
    mobilization_days : int
        Number of days it takes to mobilize the equipment.
    speed : float
        Maximum transit speed, km/hr.
    max_windspeed_transport : float
        Maximum windspeed for safe transport, m/s.
    max_windspeed_repair : float
        Maximum windspeed for safe operations, m/s.
    max_waveheight_transport : float
        Maximum waveheight for safe transport, m, default 1000 (land-based).
    max_waveheight_repair : float
        Maximum waveheight for safe operations, m, default 1000 (land-based).
    workday_start : int
        The starting hour of a workshift, in 24 hour time.
    workday_end : int
        The ending hour of a workshift, in 24 hour time.
    onsite : bool
        Indicator for if the rig and crew are based onsite.
        ... note:: if the rig and crew are onsite be sure that the start and end dates
        represent the first and last day/month of the year, respectively, and the start
        and end years represent the fist and last year in the weather file.
    method : str
        Determines if the ship will do all maximum severity repairs first or do all
        the repairs at one turbine before going to the next, by default severity.
        Should by one of "severity" or "turbine".
    max_severity : int
        Maximum severity failure that a crew can service, default None.
    """

    name: str
    equipment_rate: float
    day_rate: float
    n_day_rate: int
    hourly_rate: float
    n_hourly_rate: int
    start_month: int
    start_day: int
    start_year: int
    end_month: int
    end_day: int
    end_year: int
    capability: Union[List[str], str] = attr.ib()
    mobilization_cost: float
    mobilization_days: int
    speed: float
    max_windspeed_transport: float
    max_windspeed_repair: float
    max_waveheight_transport: float = attr.ib(default=1000)
    max_waveheight_repair: float = attr.ib(default=1000)
    workday_start: int = attr.ib(default=-1)
    workday_end: int = attr.ib(default=-1)
    onsite: bool = attr.ib(default=False)
    method: str = attr.ib(default="severity")
    max_severity: int = attr.ib(default=None)
    operating_dates: np.ndarray = attr.ib(init=False)

    @workday_start.validator
    def _check_workday_start(self, attribute, value) -> None:
        """Ensures that the workday_start is a valid time."""
        if not -1 <= value <= 24:
            raise ValueError(
                "Input 'workday_start' must be between 0 and 24, inclusive."
            )

    @workday_end.validator
    def _check_workday_end(self, attribute, value) -> None:
        """Ensures that the workday_end is a valid time."""
        if value == -1:
            return
        if not 0 <= value <= 24 or value <= self.workday_start:
            raise ValueError(
                "Input 'workday_end' must be between 0 and 24, inclusive, and after workday_start"
            )

    @capability.validator
    def _check_capability(self, attribute, value) -> None:
        """Ensures that capability has a valid input."""
        valid = set(("CTV", "SCN", "LCN", "CAB", "RMT", "DRN", "DSV"))
        values = set(convert_to_list(value, str.upper))
        invalid = values - valid
        if invalid:
            raise ValueError(f"Input 'capability' must be any combination of {valid}.")

    @method.validator
    def _check_method(self, attribute, value) -> None:
        """Ensures that method is a valid input."""
        valid = ("turbine", "severity")
        if value.lower() not in valid:
            raise ValueError(f"Input 'method' must be one of {valid}.")

    def create_date_range(self) -> np.ndarray:
        """Creates an `np.ndarray` of valid operational dates for the service equipment."""
        start_date = datetime.datetime(
            self.start_year, self.start_month, self.start_day
        )
        if self.end_year - self.start_year <= 1 or self.onsite:
            end_date = datetime.datetime(self.end_year, self.end_month, self.end_day)
            date_range = pd.date_range(start_date, end_date).date
        else:
            date_range = annual_date_range(
                self.start_day,
                self.end_day,
                self.start_month,
                self.end_month,
                self.start_year,
                self.end_year,
            )
        return date_range

    def _set_environment_shift(self, start: int, end: int) -> None:
        """Used to set the `workday_start` and `workday_end` to the environment's values.

        Parameters
        ----------
        start : int
            Starting hour of a workshift.
        end : int
            Ending hour of a workshift.
        """
        object.__setattr__(self, "workday_start", start)
        object.__setattr__(self, "workday_end", end)

    def __attrs_post_init__(self) -> None:
        object.__setattr__(
            self, "capability", convert_to_list(self.capability, str.upper)
        )
        object.__setattr__(self, "method", self.method.lower())
        object.__setattr__(self, "operating_dates", self.create_date_range())


@attr.s(frozen=True, auto_attribs=True)
class FixedCosts:
    """The fixed costs for operating a windfarm. All values are assumed to be in $/kW/yr.

    Parameters
    ----------
    operations : float
        Non-maintenance costs of operating the project.

        If a value is provided for this attribute, then it will zero out all other
        values, otherwise it will be set to the sum of the remaining values.
    operations_management_administration: float
        Activities necessary to forecast, dispatch, sell, and manage the production of
        power from the plant. Includes both the onsite and offsite personnel, software,
        and equipment to coordinate high voltage equipment, switching, port activities,
        and marine activities.

        This should only be used when not breaking down the cost into the following
        categories: `project_management_administration`,
        ``operation_management_administration``, ``marine_management``, and/or
        ``weather_forecasting``

    project_management_administration : float
        Financial reporting, public relations, procurement, parts and stock management,
        H&SE management, training, subcontracts, and general administration.
    marine_management : float
        Coordination of port equipment, vessels, and personnel to carry out inspections
        and maintenance of generation and transmission equipment.
    weather_forecasting : float
        Daily forecast of metocean conditions used to plan maintenance visits and
        estimate project power production.
    condition_monitoring : float
        Monitoring of SCADA data from wind turbine components to optimize performance
        and identify component faults.
    operating_facilities : float
        Co-located offices, parts store, quayside facilities, helipad, refueling
        facilities, hanger (if necesssary), etc.
    environmental_health_safety_monitoring : float
        Coordination and monitoring to ensure comp0liance with HSE requirements during
        operations.
    insurance : float
        Insurance policies during operational period including All Risk Property,
        Buisness Interuption, Third Party Liability, and Brokers Fee, and Storm Coverage.

        This should only be used when not breaking down the cost into the following
        categories: ``brokers_fee``, ``operations_all_risk``, ``business_interruption``,
        ``third_party_liability``, and/or ``storm_coverage``
    brokers_fee : float
        Fees for arranging the insurance package.
    operations_all_risk : float
        All Risk Property (physical damage). Sudden and unforseen physical loss or
        physical damage to teh plant/assets during the operational phase of a project.
    business_interruption : float
        Sudden and unforseen loss or physical damage to the plant/assets during the
        operational phase of a project causing an interruption.
    third_party_liability : float
        Liability imposed by law, and/or Express Contractual Liability, for bodily
        injury or property damage.
    storm_coverage : float
        Coverage from huricane and tropical storm events (tyipcally for Atlantic Coast
        projects).
    annual_leases_fees : float
        Ongoing payments, including but not limited to: payments to regulatory body for
        permission to operate at project site (terms defined within lease); payments to
        Transmission Systems Operators or Transmission Asseet Owners for rights to
        transport generated power.

        This should only be used when not breaking down the cost into the following
        categories: ``submerge_land_lease_costs`` and/or ``transmission_charges_rights``

    submerge_land_lease_costs : float
        Payments to submerged land owners for rights to build project during operations.
    transmission_charges_rights : float
        Any payments to Transmissions Systems Operators or Transmission Asset Owners for
        rights to transport generated power.
    onshore_electrical_maintenance : float
        Inspections of cables, transformer, switch gears, power compensation equipment,
        etc. and infrequent repairs

        This should only be used if not modeling these as processes within the model.
        Currently, onshore modeling is not included.
    labor : float
        The costs associated with labor, if not being modeled through the simulated
        processes.
    """

    # Total Cost
    operations: float = attr.ib(default=0)

    # Operations, Management, and General Administration
    operations_management_administration: float = attr.ib(default=0)
    project_management_administration: float = attr.ib(default=0)
    marine_management: float = attr.ib(default=0)
    weather_forecasting: float = attr.ib(default=0)
    condition_monitoring: float = attr.ib(default=0)

    operating_facilities: float = attr.ib(default=0)
    environmental_health_safety_monitoring: float = attr.ib(default=0)

    # Insurance
    insurance: float = attr.ib(default=0)
    brokers_fee: float = attr.ib(default=0)
    operations_all_risk: float = attr.ib(default=0)
    business_interruption: float = attr.ib(default=0)
    third_party_liability: float = attr.ib(default=0)
    storm_coverage: float = attr.ib(default=0)

    # Annual Leases and Fees
    annual_leases_fees: float = attr.ib(default=0)
    submerge_land_lease_costs: float = attr.ib(default=0)
    transmission_charges_rights: float = attr.ib(default=0)

    onshore_electrical_maintenance: float = attr.ib(default=0)

    labor: float = attr.ib(default=0)

    resolution: dict = attr.ib(init=False)
    hierarchy: dict = attr.ib(init=False)

    def cost_category_validator(self, name: str, sub_names: List[str]):
        """Either updates the higher-level cost to be the sum of the category's
        lower-level costs or uses a supplied higher-level cost and zeros out the
        lower-level costs.

        Parameters
        ----------
        name : str
            The higher-level cost category's name.
        sub_names : List[str]
            The lower-level cost names associated with the higher-level category.
        """
        if getattr(self, name) <= 0:
            object.__setattr__(self, name, fsum(getattr(self, el) for el in sub_names))
        else:
            for cost in sub_names:
                object.__setattr__(self, cost, 0)

    def __attrs_post_init__(self):
        grouped_names = {
            "operations_management_administration": [
                "project_management_administration",
                "marine_management",
                "weather_forecasting",
                "condition_monitoring",
            ],
            "insurance": [
                "brokers_fee",
                "operations_all_risk",
                "business_interruption",
                "third_party_liability",
                "storm_coverage",
            ],
            "annual_leases_fees": [
                "submerge_land_lease_costs",
                "transmission_charges_rights",
            ],
        }
        individual_names = [
            "operating_facilities",
            "environmental_health_safety_monitoring",
            "onshore_electrical_maintenance",
            "labor",
        ]

        # Check each category for aggregation
        for _name, _sub_names in grouped_names.items():
            self.cost_category_validator(_name, _sub_names)

        # Check for single value aggregation
        self.cost_category_validator("operations", [*grouped_names] + individual_names)

        # Create the value resolution mapping
        object.__setattr__(
            self,
            "resolution",
            {
                "low": ["operations"],
                "medium": [
                    "operations_management_administration",
                    "insurance",
                    "annual_leases_fees",
                    "operating_facilities",
                    "environmental_health_safety_monitoring",
                    "onshore_electrical_maintenance",
                    "labor",
                ],
                "high": [
                    "project_management_administration",
                    "marine_management",
                    "weather_forecasting",
                    "condition_monitoring",
                    "brokers_fee",
                    "operations_all_risk",
                    "business_interruption",
                    "third_party_liability",
                    "storm_coverage",
                    "submerge_land_lease_costs",
                    "transmission_charges_rights",
                    "operating_facilities",
                    "environmental_health_safety_monitoring",
                    "onshore_electrical_maintenance",
                    "labor",
                ],
            },
        )

        object.__setattr__(
            self,
            "hierarchy",
            {
                "operations": {
                    "operations_management_administration": [
                        "project_management_administration",
                        "marine_management",
                        "weather_forecasting",
                        "condition_monitoring",
                    ],
                    "insurance": [
                        "brokers_fee",
                        "operations_all_risk",
                        "business_interruption",
                        "third_party_liability",
                        "storm_coverage",
                    ],
                    "annual_leases_fees": [
                        "submerge_land_lease_costs",
                        "transmission_charges_rights",
                    ],
                    "operating_facilities": [],
                    "environmental_health_safety_monitoring": [],
                    "onshore_electrical_maintenance": [],
                    "labor": [],
                }
            },
        )



""""Creates the Cable class."""


from typing import Generator, List  # type: ignore

import numpy as np  # type: ignore
import simpy  # type: ignore

from wombat.core import (
    Failure,
    Maintenance,
    RepairRequest,
    SubassemblyData,
    WombatEnvironment,
)


# TODO: Need a better method for checking if a repair has been made to bring the
# subassembly back online
HOURS = 8760
TIMEOUT = 24  # Wait time of 1 day for replacement to occur


class Cable:
    """The cable system/asset class.

    Parameters
    ----------
    windfarm : `wombat.windfarm.Windfarm`
        The `Windfarm` object.
    env : WombatEnvironment
        The simulation environment.
    cable_id : str
        The unique identifier for the cable.
    start_node : str
        The starting point (`system.id`) (turbine or substation) of the cable segment.
    upstream_nodes : List[str]
        The list of upstream system ids (`system.id`) that rely on the cable segment.
    cable_data : dict
        The dictionary defining the cable segment.
    """

    def __init__(
        self,
        windfarm,
        env: WombatEnvironment,
        cable_id: str,
        start_node: str,
        upstream_nodes: List[str],
        cable_data: dict,
    ) -> None:
        """Initializes the `Cable` class.

        Parameters
        ----------
        windfarm : `wombat.windfarm.Windfarm`
            The `Windfarm` object.
        env : WombatEnvironment
            The simulation environment.
        cable_id : str
            The unique identifier for the cable.
        start_node : str
            The starting point (`system.id`) (turbine or substation) of the cable segment.
        upstream_nodes : List[str]
            The list of upstream system ids (`system.id`) that rely on the cable segment.
        cable_data : dict
            The dictionary defining the cable segment.
        """

        self.env = env
        self.windfarm = windfarm
        self.id = cable_id
        # TODO: need to be able to handle substations, which are not being modeled currently
        self.turbine = windfarm.graph.nodes(data=True)[start_node][
            "system"
        ]  # MAKE THIS START
        self.upstream_nodes = upstream_nodes

        cable_data = {**cable_data, "system_value": self.turbine.value}
        self.data = SubassemblyData(**cable_data)
        self.name = self.data.name

        self.downstream_failure = False
        self.operating_level = 1.0
        self.broken = False

        # TODO: need to get the time scale of a distribution like this
        self.processes = dict(self._create_processes())

    def _create_processes(self):
        """Creates the processes for each of the failure and maintenance types.

        Yields
        -------
        Tuple[Union[str, int], simpy.events.Process]
            Creates a dictionary to keep track of the running processes within the
            subassembly.
        """
        for level, failure in self.data.failures.items():
            yield level, self.env.process(self.run_single_failure(failure))

        for i, maintenance in enumerate(self.data.maintenance):
            yield f"m{i}", self.env.process(self.run_single_maintenance(maintenance))

    @staticmethod
    def interrupt_subassembly_processes(subassembly) -> None:
        """Interrupts all of the running processes within the subassembly except for the
        process associated with failure that triggers the catastrophic failure.

        Parameters
        ----------
        subassembly : Subassembly
            The subassembly that should have all processes interrupted.
        """
        for _, process in subassembly.processes.items():
            try:
                process.interrupt()
            except RuntimeError:
                # This error occurs for the process halting all other processes.
                pass

    def interrupt_all_subassembly_processes(self) -> None:
        """Interrupts the running processes in all of the subassemblies within `turbine`.

        Parameters
        ----------
        failure : Failure
            The failure that triggers the turbine shutting down.
        """
        for subassembly in self.turbine.subassemblies:
            self.interrupt_subassembly_processes(subassembly)

    def stop_all_upstream_processes(self, failure: Failure) -> None:
        """Stops all upstream turbines from producing power by setting their
        `System.cable_failure` to `True`.

        Parameters
        ----------
        failure : Failre
            The `Failure` that is causing a string shutdown.
        """
        for i, system_id in enumerate(self.upstream_nodes):
            node = self.windfarm.graph.nodes[system_id]["system"]

            # Check the cable to see if it's still operating first and break the
            # loop if it's already not operating
            if i > 0:
                upstream_cable = self.windfarm.graph[start_id][system_id][  # noqa: F821
                    "cable"
                ]
                if upstream_cable.downstream_failure:
                    break

                upstream_cable.downstream_failure = True
                self.env.log_action(
                    system_id="",
                    system_name="",
                    part_id=upstream_cable.id,
                    part_name=upstream_cable.name,
                    system_ol=np.nan,
                    part_ol=upstream_cable.operating_level,
                    agent=self.name,
                    action="repair request",
                    reason=failure.description,
                    additional="cable failure shutting off all upstream cables and turbines that are still operating",
                    duration=0,
                    request_id=failure.request_id,
                )

            if not node.cable_failure or node.operating_level == 0:
                node.cable_failure = True
                self.env.log_action(
                    system_id=system_id,
                    system_name=node.name,
                    part_id="",
                    part_name="",
                    system_ol=node.operating_level,
                    part_ol=np.nan,
                    agent=self.name,
                    action="repair request",
                    reason=failure.description,
                    additional="cable failure shutting off all upstream cables and turbines that are still operating",
                    duration=0,
                    request_id=failure.request_id,
                )
                start_id: str = system_id  # noqa: F841

    def run_single_maintenance(self, maintenance: Maintenance) -> Generator:
        """Runs a process to trigger one type of maintenance request throughout the simulation.

        Parameters
        ----------
        maintenance : Maintenance
            A maintenance category.

        Yields
        -------
        simpy.events.Timeout
            Time between maintenance requests.
        """
        while True:
            hours_to_next = maintenance.frequency
            if hours_to_next == 0:
                remainder = self.env.max_run_time - self.env.now
                try:
                    self.env.log_action(
                        system_id=self.turbine.id,
                        system_name=self.turbine.name,
                        part_id=self.id,
                        part_name=self.name,
                        system_ol=self.turbine.operating_level,
                        part_ol=self.operating_level,
                        agent=self.name,
                        action="none",
                        reason=f"{self.name} is not modeled",
                        additional="no maintenance will be modeled for select cable section",
                        duration=remainder,
                    )
                    yield self.env.timeout(remainder)
                except simpy.Interrupt:
                    remainder -= self.env.now

            while hours_to_next > 0:
                try:
                    # If the replacement has not been completed, then wait another minute
                    if self.operating_level == 0 or self.downstream_failure:
                        yield self.env.timeout(TIMEOUT)
                        continue

                    start = self.env.now
                    yield self.env.timeout(hours_to_next)
                    hours_to_next = 0

                    # Automatically submit a repair request
                    repair_request = RepairRequest(
                        self.turbine.id,
                        self.turbine.name,
                        self.id,
                        self.name,
                        0,
                        maintenance,
                        cable=True,
                        upstream_turbines=self.upstream_nodes,
                    )
                    repair_request = self.turbine.repair_manager.submit_request(
                        repair_request
                    )
                    self.env.log_action(
                        system_id=self.turbine.id,
                        system_name=self.turbine.name,
                        part_id=self.id,
                        part_name=self.name,
                        system_ol=self.turbine.operating_level,
                        part_ol=self.operating_level,
                        agent=self.name,
                        action="maintenance request",
                        reason=maintenance.description,
                        additional="request",
                        duration=0,
                        request_id=repair_request.request_id,
                    )

                except simpy.Interrupt:
                    if self.broken:
                        # The subassembly had so restart the maintenance cycle
                        hours_to_next = 0
                        continue
                    else:
                        # A different subassembly failed so we need to subtract the
                        # amount of passed time
                        hours_to_next -= self.env.now - start

    def run_single_failure(self, failure: Failure) -> Generator:
        """Runs a process to trigger one type of failure repair request throughout the simulation.

        Parameters
        ----------
        failure : Failure
            A failure classification.

        Yields
        -------
        simpy.events.Timeout
            Time between failure events that need to request a repair.
        """
        while True:
            hours_to_next = failure.hours_to_next_failure()
            if hours_to_next is None:
                remainder = self.env.max_run_time - self.env.now
                try:
                    self.env.log_action(
                        system_id=self.turbine.id,
                        system_name=self.turbine.name,
                        part_id=self.id,
                        part_name=self.name,
                        system_ol=self.turbine.operating_level,
                        part_ol=self.operating_level,
                        agent=self.name,
                        action="none",
                        reason=f"{self.name} is not modeled",
                        additional="no failures will be modeled for this cable section",
                        duration=remainder,
                    )
                    yield self.env.timeout(remainder)
                except simpy.Interrupt:
                    remainder -= self.env.now

            while hours_to_next > 0:  # type: ignore
                try:
                    if self.operating_level == 0 or self.downstream_failure:
                        yield self.env.timeout(TIMEOUT)
                        continue

                    start = self.env.now
                    yield self.env.timeout(hours_to_next)
                    hours_to_next = 0
                    self.operating_level *= 1 - failure.operation_reduction

                    if failure.operation_reduction == 1:
                        self.broken = True
                        self.interrupt_all_subassembly_processes()

                        # Remove previously submitted requests as a replacement is required
                        _ = self.turbine.repair_manager.purge_subassembly_requests(
                            self.turbine.id, self.id
                        )

                    # Automatically submit a repair request
                    repair_request = RepairRequest(
                        self.turbine.id,
                        self.turbine.name,
                        self.id,
                        self.name,
                        failure.level,
                        failure,
                        cable=True,
                        upstream_turbines=self.upstream_nodes,
                    )
                    repair_request = self.turbine.repair_manager.submit_request(
                        repair_request
                    )

                    self.env.log_action(
                        system_id=self.turbine.id,
                        system_name=self.turbine.name,
                        part_id=self.id,
                        part_name=self.name,
                        system_ol=self.turbine.operating_level,
                        part_ol=self.operating_level,
                        agent=self.name,
                        action="repair request",
                        reason=failure.description,
                        additional=f"severity level {failure.level}",
                        duration=0,
                        request_id=repair_request.request_id,
                    )

                except simpy.Interrupt:
                    if self.broken:
                        # The subassembly had to be replaced so the timing to next failure
                        # will reset
                        hours_to_next = 0
                        continue
                    else:
                        # A different subassembly failed so we need to subtract the
                        # amount of passed time
                        hours_to_next -= self.env.now - start
